#!/usr/bin/python
# -*- coding: utf-8 -*-
import gtk, gtk.glade, gobject
import os, sys, time, math

class bpmTap:
	def __init__(self):
		# Load Glade
		gladefile = "bpm-tap.glade"
		self._windowname = "windowBPMtap"
		self.widgets = gtk.glade.XML(gladefile, self._windowname)

		signals = {
			"onButtonClick" : self.onButtonClick,
			"onButtonTap" : self.onWindowTap,
			"onButtonRelease" : self.onWindowKeyRelease,
			"onWindowTap" : self.onWindowTap,
			"onWindowKeyRelease" : self.onWindowKeyRelease,
			"onQuit" : self.onDestroy,
			"plotGraph" : self.plotGraph,
		}

		self.widgets.signal_autoconnect(signals)

		self.window = self.widgets.get_widget(self._windowname)
		self.bpmentry = self.widgets.get_widget("entryBPM")
		self.statsrawlabel = self.widgets.get_widget("labelStatsRaw")
		self.statsfloatlabel = self.widgets.get_widget("labelStatsFloating")
		self.plotwidget = self.widgets.get_widget("drawingarea")
		self.flasherwidget1 = self.widgets.get_widget("imageFlasher1")
		self.flasherwidget2 = self.widgets.get_widget("imageFlasher2")

		self._stuckkey = None

		self.curbpm = 0.0
		self.clicks = 0
		self.last = 0
		self.lasttimes = []
		self.lastbpms = []
		self.lastfloat = []
		self.lastfloatsq = []
		self._sum = 0.0
		self._sqsum = 0.0
		self._avgcnt = 0
		self._min = 0
		self._max = 0
		self.floatbpm = 0.0
		self.floatsq = 0.0
		self.keep = 100

		self.pause = 3
		self.minweight = 0.01

		self._nextflash = 0
		self._flashid = None
		self._flashState = False

	def onDestroy(self, *args):
		gtk.main_quit()

	def countTap(self, now):
		now = now / 1000.
		# reset?
		if now - self.last > self.pause:
			self.clicks = 0
			self.curbpm = 0.0

			self.lasttimes = []
			self.lastbpms = []
			self.lastfloat = []
			self.lastfloatsq = []
			self._sum = 0.0
			self._sqsum = 0.0
			self._avgcnt = 0
			self._min = 0
			self._max = 0
			self.floatbpm = 0.0
			self.floatsq = 0.0
		elif now <= self.last:
			print >>sys.stderr, "Bad timing of events: ", now, self.last
		else:
			# Use prev 5 values to calc BPM averge
			bpms = []
			bpms.append(60.0 / (now - self.last))
			# and four out of the list
			for i in xrange(1,5):
				if len(self.lasttimes) <= i: break
				bpms.append( (i+1) * 60.0 / (now - self.lasttimes[-i]))
			# remove one error - skipped beat?
			#if len(bpms) > 3:
			#	worst = -1
			#	worstdelta = 0
			#	for i in xrange(len(bpms)):
			#		if abs(bpms[i] - self.curbpm) > worstdelta:
			#			worst = i
			#			worstdelta = abs(bpms[i] - self.curbpm)
			#	if worst > -1:
			#		del bpms[worst]
			# average
			self.curbpm = reduce(lambda x,y: x+y, bpms) / len(bpms)

			# exponentially weighted floating average
			weight = (1.0 / self.clicks) ** .5
			if weight < self.minweight: weight = self.minweight
			self.floatbpm = self.floatbpm * (1.0 - weight) + self.curbpm * weight
			self.floatsq  = self.floatsq * (1.0 - weight) + (self.curbpm * self.curbpm) * weight

			if self.curbpm < self._min or self._avgcnt == 0: self._min = self.curbpm
			if self.curbpm > self._max or self._avgcnt == 0: self._max = self.curbpm
			self._sum += self.curbpm
			self._sqsum += self.curbpm * self.curbpm
			self._avgcnt += 1

			# update history
			self.lasttimes = self.lasttimes[-(self.keep-1):] + [self.last]
			self.lastbpms = self.lastbpms[-(self.keep-1):] + [self.curbpm]
			self.lastfloat = self.lastfloat[-(self.keep-1):] + [self.floatbpm]
			self.lastfloatsq = self.lastfloatsq[-(self.keep-1):] + [self.floatsq]

			#print self.curbpm, (self._sum / self._avgcnt), self.floatbpm

		self.last = now
		self.clicks += 1

		self.scheduleFlash()

		gobject.idle_add(self.updateCounter)
		self.plotwidget.window.invalidate_rect(None, True)

	def updateCounter(self):
		self.bpmentry.set_text("%.1f" % self.floatbpm)

		if self.clicks > 0 and self._avgcnt >= 2:
			#avgest = self._sum / self._avgcnt
			#stddev = math.sqrt( (self._sqsum - (self._sum * self._sum) / self._avgcnt) / (self._avgcnt - 1))
			try:
				floatstddev = math.sqrt( self.floatsq - self.floatbpm * self.floatbpm )
			except ValueError, e:
				floatstddev = 0.

			self.statsrawlabel.set_markup("Clicks: <b>%d</b>\nRaw BPM: <b>%.1f</b>" % (self.clicks, self.curbpm))
			self.statsfloatlabel.set_markup("Floating BPM estimation:\n<b>%.1f ± %.2f</b>" % (self.floatbpm, floatstddev))
		else:
			self.statsrawlabel.set_markup("Clicks: <b>%d</b>\nRaw BPM: <b>n/a</b>" % (self.clicks))
			self.statsfloatlabel.set_markup("Floating BPM estimation:\n<b>n/a</b>")

	
	def onWindowTap(self, widget, event):
		# ignore a "stuck" key.
		if self._stuckkey == event.keyval: return
		self._stuckkey = event.keyval
		# ignore keys with modifiers
		if event.state & gtk.gdk.CONTROL_MASK: return
		if event.state & gtk.gdk.MOD1_MASK: return

		# ignore modifier touches - hackish
		# TODO: cleaner way of detecting these?
		# "deadkeys"?
		keyname = gtk.gdk.keyval_name(event.keyval)
		if keyname.find("Alt") >= 0: return
		if keyname.find("Control") >= 0: return
		if keyname.find("Shift") >= 0: return
		if keyname.find("Super") >= 0: return
		if keyname.find("Meta") >= 0: return
		
		self.countTap(event.time)

	def onWindowKeyRelease(self, widget, event):
		self._stuckkey = None

	def onButtonClick(self, widget, event):
		#print "Click.", event
		if event.type == gtk.gdk.BUTTON_PRESS:
			self.countTap(event.time)

	def plotGraph(self, widget, event):
		# Create the cairo context
		cr = self.plotwidget.window.cairo_create()
		cr.rectangle(event.area.x, event.area.y, event.area.width, event.area.height)
		cr.clip()

		(width, height) = self.plotwidget.window.get_size()

		# background
		cr.set_source_rgb(1, 1, 1)
		cr.rectangle(0, 0, width, height)
		cr.fill()

		# get min/max/avg/stddev BPM
		if self._avgcnt < 2: return
		bpmmin = self._min * 0.95
		bpmmax = self._max * 1.05
		avgest = self._sum / self._avgcnt
		stddev = math.sqrt( (self._sqsum - (self._sum * self._sum) / self._avgcnt) / (self._avgcnt - 1))

		timemin = self.lasttimes[0]
		timemax = self.lasttimes[-1]
		def relx(v):
			return (v - timemin) / (timemax - timemin) * width

		def rely(v):
			return (bpmmax - v) / (bpmmax - bpmmin) * height

		# draw the floating averages
		cr.set_line_width(2)
		cr.set_source_rgb(0, 1, 0)
		cr.move_to(0, rely(self.lastfloat[0]))
		for i in xrange(1,len(self.lastfloat)):
			cr.line_to(relx(self.lasttimes[i]), rely(self.lastfloat[i]))
		cr.stroke()

		# floating error lines
		cr.set_line_width(1)
		cr.set_source_rgb(.4, 1, .4)
		cr.move_to(0, rely(self.lastfloat[0]))
		for i in xrange(0,len(self.lastfloat)):
			err = math.sqrt(abs(self.lastfloatsq[i] - self.lastfloat[i] * self.lastfloat[i]))
			cr.line_to(relx(self.lasttimes[i]), rely(self.lastfloat[i] + err))
		cr.stroke()
		cr.move_to(0, rely(self.lastfloat[0]))
		for i in xrange(1,len(self.lastfloat)):
			err = math.sqrt(abs(self.lastfloatsq[i] - self.lastfloat[i] * self.lastfloat[i]))
			cr.line_to(relx(self.lasttimes[i]), rely(self.lastfloat[i] - err))
		cr.stroke()

		# draw the history line
		cr.set_line_width(1)
		cr.set_source_rgb(0.0, 0.0, 0.8)
		cr.move_to(0, rely(self.lastbpms[0]))
		for i in xrange(1,len(self.lastbpms)):
			cr.line_to(relx(self.lasttimes[i]), rely(self.lastbpms[i]))
		cr.stroke()

	def start(self):
		self.window.show()

	def scheduleFlash(self):
		if self.floatbpm > 20 and self.floatbpm < 500:
			scheduledflash = self._nextflash
			interval = int(60000. / self.floatbpm)
			self._nextflash = time.time() + interval/1000.

			# remove flashes when they 'conflict' with the now scheduled flash.
			if self._flashid and scheduledflash:
				if abs(scheduledflash - time.time()) > abs(self._nextflash - scheduledflash):
					gobject.source_remove(self._flashid)
			self._flashid = gobject.timeout_add(interval, self.flash, self._nextflash)

	def flash(self, mytime):
		self._flashState = not self._flashState
		if self._flashState:
			self.flasherwidget1.set_from_stock(gtk.STOCK_NO, gtk.ICON_SIZE_SMALL_TOOLBAR)
			self.flasherwidget2.set_from_stock(gtk.STOCK_YES, gtk.ICON_SIZE_SMALL_TOOLBAR)
		else:
			self.flasherwidget1.set_from_stock(gtk.STOCK_YES, gtk.ICON_SIZE_SMALL_TOOLBAR)
			self.flasherwidget2.set_from_stock(gtk.STOCK_NO, gtk.ICON_SIZE_SMALL_TOOLBAR)

		if mytime == self._nextflash and self.floatbpm > 20 and self.floatbpm < 500:
			interval = int(60000. / self.floatbpm)
			self._nextflash = time.time() + interval/1000.
			self._flashid = gobject.timeout_add(interval, self.flash, self._nextflash)
		# stop old timeout
		return False

if __name__ == "__main__":
        ui = bpmTap()
        gtk.gdk.threads_init()
        ui.start()
        gtk.main()
